<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use function defined;
use Infomaker\Everyware\Support\Storage\Cache;
use SplFileInfo;
use Twig\Error\Error;
use Twig\Loader\FilesystemLoader as TwigFilesystemLoader;

/**
 * FilesystemLoader
 *
 * @link    http://infomaker.se
 * @package Infomaker\Everyware\Twig
 * @since   Infomaker\Everyware\Twig\FilesystemLoader 0.1
 */
class FilesystemLoader extends TwigFilesystemLoader
{
    /**
     * Add suffix so we don't have to delete it in deploy every time
     *
     * @var string
     */
    private $cacheKey = 'everyware_twig_paths_cache';

    /**
     * @var FilesystemLoader
     */
    private static $instance;
    /**
     * Time until expiration in seconds. Set to one Week: (60 * 60 * 24 * 7)
     * @var int
     */
    private $pathCacheExpire = 604800;

    /**
     * Create new instance if none exist
     *
     * @return FilesystemLoader
     */
    public static function getInstance(): FilesystemLoader
    {
        if (static::$instance === null) {
            static::$instance = new static;
        }

        return static::$instance;
    }

    /**
     * Adds the folders containing the files to be loaded
     *
     * @param string $folderPath
     * @param string $namespace
     * @param array  $exclude
     *
     * @return void
     * @throws Error
     */
    public function addPathAndSubfolders(
        string $folderPath,
        string $namespace = self::MAIN_NAMESPACE,
        array $exclude = []
    ): void {

        if ( ! file_exists($folderPath)) {
            throw new Error("{$folderPath} does not exist");
        }

        $cacheKey = $this->getPathCacheKey($folderPath);
        $paths = $this->getCache($cacheKey);

        if (empty($paths)) {
            $dir = PathFilter::setupFromPath($folderPath, $exclude);
            $paths = [];
            foreach ($dir->filtered() as $d) {
                if ($d instanceof SplFileInfo) {
                    $paths[] = str_replace('//', '/', $d->getPath());
                }
            }

            $this->setCache($cacheKey, $paths);
            $paths = array_unique($paths);
        }

        foreach ($paths as $path) {
            $this->addPath($path, $namespace);
        }
    }

    /**
     * @param string $cacheKey
     *
     * @return array
     */
    protected function getCache(string $cacheKey): array
    {
        return Cache::get($cacheKey, []);
    }

    /**
     * @param string $cacheKey
     * @param array  $paths
     *
     * @return void
     */
    protected function setCache(string $cacheKey, array $paths): void
    {
        Cache::set($cacheKey, $paths, $this->pathCacheExpire);
    }

    /**
     * @param string $folderPath
     *
     * @return string
     */
    protected function getPathCacheKey(string $folderPath): string
    {
        $cacheKey = $this->cacheKey . '_' . md5($folderPath);

        // Add commit to invalidate cache from last release
        if (defined('GIT_COMMIT')) {
            $cacheKey .= '_' . GIT_COMMIT;
        }

        return $cacheKey;
    }
}
