<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use Infomaker\Everyware\Support\Device;
use Infomaker\Everyware\Support\NewRelicLog;
use Infomaker\Everyware\Support\Str;
use RuntimeException;
use Twig\Environment;
use Twig\Error\Error;
use Twig\Lexer;
use Twig\Loader\LoaderInterface;
use Twig\Parser;
use Twig\TwigFilter;
use Twig\TwigFunction;


class ViewSetup
{
    /** @var int */
    protected $cacheTime = 86400;

    /** @var bool */
    protected $cache;

    /** @var bool */
    protected $debug;

    /**
     * Contains functions that will be available in the templates.
     * If the same function name is added several times only the last one is kept.
     * @var array
     */
    protected $functions = [];

    /**
     * Contains filters that will be available in the templates.
     * If the same filters name is added several times only the last one is kept.
     * @var array
     */
    protected $filters = [];
    /**
     * Contains filters that will be available in the templates.
     * @var array
     */
    protected $globals = [];

    /**
     * Contains the FilesystemLoader aka all template-paths available in the templates.
     * @var LoaderInterface
     */
    protected $loader;
    protected $everyTwig;

    private static $instance;
    /**
     * @var Environment
     */
    private $twigEnvironment;

    private function __construct(EveryTwig $everyTwig, LoaderInterface $loader)
    {
        //Add global device variable
        $this->addGlobal('device', Device::currentDevice());
        $this->addTwigFunctions();

        $this->cache = ! is_dev();
        $this->debug = is_dev();

        $this->setLoader($loader);
        $this->setEveryTwig($everyTwig);
    }

    /**
     * Create new instance if none exist
     *
     * @return ViewSetup
     */
    public static function getInstance(): ViewSetup
    {
        if (static::$instance === null) {
            static::$instance = new static(EveryTwig::getInstance(), FilesystemLoader::getInstance());
        }

        return static::$instance;
    }

    public function setLoader(LoaderInterface $loader): void
    {
        $this->loader = $loader;
    }

    public function setEveryTwig(EveryTwig $everyTwig): void
    {
        $this->everyTwig = $everyTwig;
    }

    private function getLoader(): LoaderInterface
    {
        return $this->loader;
    }

    public function getEveryTwig(): EveryTwig
    {
        return $this->everyTwig;
    }

    public function addFilter(string $name, $callable = null, array $options = []): self
    {
        $filter = new TwigFilter($name, $callable, $options);

        // If environment has yet to be initialized we'll store it for when it is
        if ($this->environmentInitialized()) {
            $this->twigEnvironment->addFilter($filter);
        } else {
            $this->storeFilter($name, $filter);
        }

        return $this;
    }

    public function addFunction(string $name, $callable = null, array $options = [])
    {
        $function = new TwigFunction($name, $callable, $options);

        // If environment has yet to be initialized we'll store it for when it is
        if ($this->environmentInitialized()) {
            $this->twigEnvironment->addFunction($function);
        } else {
            $this->storeFunction($name, $function);
        }

        return $this;
    }

    public function addGlobal(string $key, $value): self
    {
        // If environment has yet to be initialized we'll store it for when it is
        if ($this->environmentInitialized()) {
            $this->twigEnvironment->addGlobal($key, $value);
        } else {
            $this->storeGlobal($key, $value);
        }

        return $this;
    }

    public function addWpFunction(string $functionName, string $name = null): self
    {
        $name = ( ! empty($name) && $name !== null) ? $name : $functionName;
        $this->addFunction($name, $functionName);

        return $this;
    }

    public function init(array $options = []): self
    {
        if ($this->environmentInitialized()) {
            return $this;
        }

        $everyTwig = $this->getEveryTwig();

        $this->twigEnvironment = $everyTwig->init($this->getLoader(), $this->functions, $this->filters,
            array_replace([
                'debug' => $this->debug,
                'cache' => $this->cache,
                'cachetime' => $this->cacheTime,
            ], $options));

        foreach ($this->globals as $key => $value) {
            $this->twigEnvironment->addGlobal($key, $value);
        }

        $this->twigEnvironment->setLexer(new Lexer($this->twigEnvironment));
        $this->twigEnvironment->setParser(new Parser($this->twigEnvironment));

        return $this;
    }

    /**
     * @param string     $namespace
     * @param string     $folderPath
     * @param array|null $exclude
     *
     * @return ViewSetup
     */
    public function registerTwigFolder(string $namespace, string $folderPath, array $exclude = null): self
    {
        if (is_dir($folderPath)) {
            try {
                $loader = $this->getLoader();

                if ( ! $loader instanceof FilesystemLoader) {
                    throw new RuntimeException('"registerTwigFolder" is not available with registered loaders that are not instance of ' . FilesystemLoader::class);
                }

                $loader->addPathAndSubfolders($folderPath, $namespace, $exclude ?? []);
            } catch (RuntimeException $e) {
                NewRelicLog::error('Failed to register twig directory: ' . $folderPath, $e);
            } catch (Error $e) {
                NewRelicLog::error('Failed to register twig directory: ' . $folderPath, $e);
            }
        }

        return $this;
    }

    protected function addTwigFunctions(): self
    {
        // All methods from DefaultTemplateFunctions will be added to twig as global functions.
        // The method-name will be converted to snake_case to match PSR-1 for "global functions"
        // E.g. methodName => method_name
        $twigFunctions = new DefaultTemplateFunctions();
        foreach (get_class_methods($twigFunctions) as $funcName) {
            $this->addFunction(Str::snake($funcName), [$twigFunctions, $funcName]);
        }

        return $this;
    }

    private function environmentInitialized(): bool
    {
        return $this->twigEnvironment !== null;
    }

    /**
     * @param string     $name
     * @param TwigFilter $filter
     */
    private function storeFilter(string $name, TwigFilter $filter): void
    {
        $this->filters[$name] = $filter;
    }

    /**
     * @param string       $name
     * @param TwigFunction $function
     */
    private function storeFunction(string $name, TwigFunction $function): void
    {
        $this->functions[$name] = $function;
    }

    /**
     * @param string $key
     * @param mixed  $value
     */
    private function storeGlobal(string $key, $value): void
    {
        $this->globals[$key] = $value;
    }
}
